--[[
********************************
Licence: GPL
(C)2012 rudi, Version: 0.5

CORDIC LIBRARY
integer based trigonometric functions
based on http://www.andreadrian.de/c-workshop/index.html
********************************

all angels in rad or deg
all values multiplied by SCALE (eq. CORDIC_SCALE/SCALE_DIV)

ranges:
  sind(ALL) = -0.999..+0.999
  cosd(ALL) = -0.999..+0.999
  tand(ALL) = -3332.666..3332.666 (90/270 = -3332.666)
  asind(-0.986..+0.986) = -90..+90
  acosd(-0.986..+0.986) = 0..+180
  atand(-214747..+214747) = -90..+90
  recd(0..21*, ALL) = pold(-21..+21*, -21..+21*)
  pold(-10..+10*, -10..+10*) = recd(0..14.14*, 0..360)
 
  sinr 
  cosr 
  tanr
  asinr
  acosr
  atanr
  recr
  polr

  (*) greater ranges untested
]]

cordicLib = {}

cordicLib.version = "0.5"

-- intern CORDIC defines
local ROTATE, VECTOR = 0, 1 -- f_coedic
local CORDIC_SCALE = 10000  -- eqal 1 for cordic
local INV_GAIN_CIRCLE = 6073

-- atan(1), atan(0,5) ... all in RAD and DEG multiplied by CORDIC_SCALE
local atan_tab = {
    {7854, 4636, 2450, 1244, 624, 312, 156, 78, 39, 20, 10, 5, 2, 1, 1},
    {450000, 265650, 140362, 71250, 35763, 17899, 8951, 4476, 2238, 1119, 559, 279, 139, 69}
}
-- intern defines
local SCALE_DIV = 10
local ANGEL, POINT = 0, 1   -- q1 covert mode
local RAD, DEG = 1, 2       -- t_angel
local FULL_CIRCLE = {62831, 360 * CORDIC_SCALE}  -- 2*PI, 360
local HALF_CIRCLE = {FULL_CIRCLE[RAD] / 2, FULL_CIRCLE[DEG] / 2}
local QUART_CIRCLE = {FULL_CIRCLE[RAD] / 4, FULL_CIRCLE[DEG] / 4}

-- CORDIC MAIN

local function cordic_circle(t_angel, f_cordic, c, x, y, z)
    assert(t_angel and f_cordic and c and x and y and z, "nil argument in cordic")
    local xstep, ystep, zstep
    local dstep = 1
    for i = 1, #atan_tab[t_angel] do
        xstep = x / dstep
        ystep = y / dstep
        zstep = atan_tab[t_angel][i]
        if (f_cordic == VECTOR) and (y >= c) or (z < c) then
            x = x + ystep
            y = y - xstep
            z = z + zstep
        else
            x = x - ystep
            y = y + xstep
            z = z - zstep
        end
        dstep = 2 * dstep
    end
    return x, y, z
end

-- additional functions

local function scale_in(value)
    return value * SCALE_DIV
end

local function scale_out(value)
    return value / SCALE_DIV
end

local function convert_to_q1(t_angel, value, mode)
    local res, q
    if mode == ANGEL and type(value) == "number" then
        res = (value % FULL_CIRCLE[t_angel] + FULL_CIRCLE[t_angel]) % FULL_CIRCLE[t_angel]
        q = res / QUART_CIRCLE[t_angel] + 1
        if q == 2 then res = HALF_CIRCLE[t_angel] - res
        elseif q == 3 then res = res - HALF_CIRCLE[t_angel]
        elseif q == 4 then res = FULL_CIRCLE[t_angel] - res
        end
    elseif mode == POINT and type(value) == "table" then
        local cosx, sinx = (value[1] >= 0), (value[2] >= 0)
        q = 1
        if not cosx and sinx then q = 2
        elseif not cosx and not sinx then q = 3
        elseif cosx and not sinx then q = 4
        end
        res = {math.abs(value[1]), math.abs(value[2])}
    end
    return res, q
end

local function convert_from_q1(t_angel, value, mode, q)
    local res
    if mode == ANGEL and type(value) == "number" then
        res = value
        if q == 2 then res = HALF_CIRCLE[t_angel] - res
        elseif q == 3 then res = HALF_CIRCLE[t_angel] + res
        elseif q == 4 then res = FULL_CIRCLE[t_angel] - res
        end
    elseif mode == POINT and type(value) == "table" then
        local cosx, sinx = 1, 1
        if q == 2 or q == 3 then cosx = -1 end
        if q == 3 or q == 4 then sinx = -1 end
        res = {cosx * value[1], sinx * value[2]}
    end
    return res
end

-- calculate a * b / c with overflow prevention and different precision
local function integer_mul_div(a, b, c)
    local function _sign(_x)
        return (_x < 0) and -1 or 1
    end
    -- greatest common divisor with
    local function _gcd(_x, _y)
        if _x < _y then _x, _y = _y, _x end
        while _y > 0 do _x, _y = _y, _x % _y end
        return _x
    end
    -- max value = 21'47483647; _x*_y -> >8 chars or ==8 chars and > 21 -> Overflow
    local function _is_mul_overflow(_x, _y)
        if (_last_mul_x ~= _x) or (_last_mul_y ~= _y) then
            _last_mul_x, _last_mul_y = _x, _y
            local _xs, _ys = tostring(_x), tostring(_y)
            _last_mul_res = (#_xs + #_ys - 2 > 8) or (#_xs + #_ys - 2 == 8) and
                ((tonumber(string.sub(_xs, 1, 1)) + 1) * (tonumber(string.sub(_ys, 1, 1)) + 1) > 21)
        end
        return _last_mul_res
    end

    -- (_x + _y) > 2147483647 -> Overflow
    local function _is_add_overflow(_x, _y)
        return (2147483647 - _x) < _y
    end

    assert(c~=0, "division by 0")
    local _a, _b, _c, _s = math.abs(a), math.abs(b), math.abs(c), _sign(a) * _sign(b) / _sign(c)
    -- _a is always the greater or equal value
    if _a < _b then _a, _b = _b, _a end
    
    if _b == 0 then return 0
    elseif _b == 1 then return _s * _a 
    end
    
    -- overflow prevention precise part
    if _is_mul_overflow(_a, _b) then
        -- cancelling down fraction _a, _c
        local _t = _gcd(_a, _c)
        if _t > 1 then _a, _c = _a / _t, _c / _t end
    end
    if _is_mul_overflow(_a, _b) then
        -- cancelling down fraction _b, _c
        local _t = _gcd(_b, _c)
        if _t > 1 then _b, _c = _b / _t, _c / _t end
    end
    if (_a > _c) and _is_mul_overflow(_a, _b) then
        -- mixed numbers fraction
        local _a1, _a2 = integer_mul_div(_a / _c, _b, 1), integer_mul_div(_a % _c, _b, _c)
        if not _is_add_overflow(_a1, _a2) then return _s * (_a1 + _a2) end
    end
    
    --overflow prevention imprecise part
    --scale down: divide in half _c and even number from _b or the greater value _a
    if _is_mul_overflow(_a, _b)then
        while (_c > 1) and _is_mul_overflow(_a, _b) do
            _c = _c / 2
            if _b / 2 * 2 == _b then _b = _b / 2 else _a = _a / 2 end
            if _a < _b then _a, _b = _b, _a end
        end
    end

    if not _is_mul_overflow(_a, _b) then return _s * _a * _b / _c end
    return (a > b) and a / c * b or b / c * a
end

local function mul_scaled(a, b)
    return integer_mul_div(a, b, CORDIC_SCALE)
end

local function div_scaled(a, b)
    return integer_mul_div(CORDIC_SCALE, a, b)
end

--[[
calculate INV_GAIN_CIRCLE

function cordicLib.calc_cordic_gain()
    GAIN_CIRCLE = cordic_circle(ROTATE, 0, CORDIC_SCALE, 0, 0)
    INV_GAIN_CIRCLE = div_scaled(CORDIC_SCALE, GAIN_CIRCLE)   -- 1 / GAIN_CIRCLE
    print(string.format("local INV_GAIN_CIRCLE = %d", INV_GAIN_CIRCLE))
end
--]]

-- BASE CORDIC ROTATE
-- phi = ALL
local function cordic_sincos(t_angel, phi)
    local q1_phi, q = convert_to_q1(t_angel, phi, ANGEL)
    local tab = {cordic_circle(t_angel, ROTATE, 0, INV_GAIN_CIRCLE, 0, q1_phi)}
    tab = convert_from_q1(t_angel, tab, POINT, q)
    -- return sin = -0.999..+0.999, cos = -0.999..+0.999
    return tab[2], tab[1]
end

-- BASE CORDIC VECTOR
--x = -0.986..+0.986
local function cordic_asincos(t_angel, x)
    local q1_x, q = convert_to_q1(t_angel, {0, x}, POINT)
    local tab = {cordic_circle(t_angel, VECTOR, -q1_x[2], INV_GAIN_CIRCLE, 0, 0)}
    --x > 0.986 can get angel greater then 90
    if tab[3] > QUART_CIRCLE[t_angel] then tab[3] = QUART_CIRCLE[t_angel] end
    local phi = convert_from_q1(t_angel, tab[3], ANGEL, q)
    -- return phi = -90..+90
    return (q == 4) and phi - FULL_CIRCLE[t_angel] or phi
end

--x = -10..+10, y = -10..+10
local function cordic_atanhyp(t_angel, x, y)
    local q1_x, q = convert_to_q1(t_angel, {x, y}, POINT)
    local tab = {cordic_circle(t_angel, VECTOR, 0, q1_x[1], q1_x[2], 0)}
    local phi = convert_from_q1(t_angel, tab[3], ANGEL, q)
    -- return phi = -90..270, hypotenuse/INV_GAIN_CIRCLE = 0..23.2 (23.2 eq sqrt(2*10^2)/INV_GAIN_CIRCLE)
    return  (q == 4) and phi - FULL_CIRCLE[t_angel] or phi, tab[1]
end

-- intern trigonometric functions

-- CORDIC ROTATE functions
--sin(ALL) = -0.999..+0.999
local function cordic_sin(t_angel, phi)
    local tab = {cordic_sincos(t_angel, phi)}
    return tab[1]
end

--cos(ALL) = -0.999..+0.999
local function cordic_cos(t_angel, phi)
    local tab = {cordic_sincos(t_angel, phi)}
    return tab[2]
end

--tan(ALL) = -3332.666..3332.666 (90/270 = -3332.666)
local function cordic_tan(t_angel, phi)
    local tab = {cordic_sincos(t_angel, phi)}
    return div_scaled(tab[1], tab[2])
end

--rec(0..21, ALL) = pol(-21..+21, -21..+21)
local function cordic_rec(t_angel, r, theta)
    local tab = {cordic_sincos(t_angel, theta)}
    -- return x, y
    return mul_scaled(r, tab[2]), mul_scaled(r, tab[1])
end

-- CORDIC VECTOR functions

--asin(-0.98..+0.98) = -90..+90
local function cordic_asin(t_angel, x)
    return cordic_asincos(t_angel, x)
end

--acos(-0.98..+0.98) = 0..+180
local function cordic_acos(t_angel, x)
    return QUART_CIRCLE[t_angel] - cordic_asincos(t_angel, x)
end

--atan(-214747..+214747) = -90..+90
local function cordic_atan(t_angel, x)
    local tab = {cordic_atanhyp(t_angel, CORDIC_SCALE, x)}
    return tab[1]
end

--pol(-10..+10, -10..+10) = rec(0..14.14, 0..360)
local function cordic_pol(t_angel, x, y)
    local tab = {cordic_atanhyp(t_angel, x, y)}
    -- r, theta
    return mul_scaled(tab[2], INV_GAIN_CIRCLE), tab[1] < 0 and tab[1] + FULL_CIRCLE[t_angel] or tab[1]
end

-- extern define
cordicLib.SCALE = CORDIC_SCALE / SCALE_DIV
cordicLib.PI2 = scale_out(FULL_CIRCLE[RAD])
cordicLib.PI = scale_out(HALF_CIRCLE[RAD])
cordicLib.PI_2 = scale_out(QUART_CIRCLE[RAD])

-- extern convert functions
function cordicLib.rad2deg(phi)
    local _phi = scale_in(phi)
    return scale_out(integer_mul_div(_phi, FULL_CIRCLE[DEG], FULL_CIRCLE[RAD]))
end

function cordicLib.deg2rad(phi)
    local _phi = scale_in(phi)
    return scale_out(integer_mul_div(_phi, FULL_CIRCLE[RAD], FULL_CIRCLE[DEG]))
end

function cordicLib.integer_mul_div(a, b, c)
    -- = a * b / c
    return integer_mul_div(a, b, c)
end

-- extern trigonometric functions

-- angel in rad
function cordicLib.sinr(phi)
    local _phi = scale_in(phi)
    local _x = cordic_sin(RAD, _phi)
    return scale_out(_x)
end

function cordicLib.cosr(phi)
    local _phi = scale_in(phi)
    local _x = cordic_cos(RAD, _phi)
    return scale_out(_x)
end

function cordicLib.tanr(phi)
    local _phi = scale_in(phi)
    local _x = cordic_tan(RAD, _phi)
    return scale_out(_x)
end

function cordicLib.asinr(x)
    local _x = scale_in(x)
    local _phi = cordic_asin(RAD, _x)
    return scale_out(_phi)
end

function cordicLib.acosr(x)
    local _x = scale_in(x)
    local _phi = cordic_acos(RAD, _x)
    return scale_out(_phi)
end

function cordicLib.atanr(x)
    local _x = scale_in(x)
    local _phi = cordic_atan(RAD, _x)
    return scale_out(_phi)
end

function cordicLib.recr(r, theta)
    local _r, _theta = scale_in(r), scale_in(theta)
    local _x, _y = cordic_rec(RAD, _r, _theta)
    return scale_out(_x), scale_out(_y)
end

function cordicLib.polr(x, y)
    local _x, _y = scale_in(x), scale_in(y)
    local _r, _theta = cordic_pol(RAD, _x, _y)
    return scale_out(_r), scale_out(_theta)
end

-- angel in deg
function cordicLib.sind(phi)
    local _phi = scale_in(phi)
    local _x = cordic_sin(DEG, _phi)
    return scale_out(_x)
end

function cordicLib.cosd(phi)
    local _phi = scale_in(phi)
    local _x = cordic_cos(DEG, _phi)
    return scale_out(_x)
end

function cordicLib.tand(phi)
    local _phi = scale_in(phi)
    local _x = cordic_tan(DEG, _phi)
    return scale_out(_x)
end

function cordicLib.asind(x)
    local _x = scale_in(x)
    local _phi = cordic_asin(DEG, _x)
    return scale_out(_phi)
end

function cordicLib.acosd(x)
    local _x = scale_in(x)
    local _phi = cordic_acos(DEG, _x)
    return scale_out(_phi)
end

function cordicLib.atand(x)
    local _x = scale_in(x)
    local _phi = cordic_atan(DEG, _x)
    return scale_out(_phi)
end

function cordicLib.recd(r, theta)
    local _r, _theta = scale_in(r), scale_in(theta)
    local _x, _y = cordic_rec(DEG, _r, _theta)
    return scale_out(_x), scale_out(_y)
end

function cordicLib.pold(x, y)
    local _x, _y = scale_in(x), scale_in(y)
    local _r, _theta = cordic_pol(DEG, _x, _y)
    return scale_out(_r), scale_out(_theta)
end

return cordicLib
